/******************************************************************************

Floppy disk emulator using SD card

Emulates the Altair Floppy Disk System consisting of the
MITs 88-DCDD floppy disk controller and 88-Disk floppy drives

2014/04/06 - Version 2.0 - Bug fixes

*****************************************************************************/

#include <p24FJ128GA106.h>
#include "FSIO.h"			//Memory disk drive file system
#define USE_AND_OR
#include "PPS.h"			//Peripheral Pin Select
#include<ports.h>

    _CONFIG2(IESO_OFF & FNOSC_FRCPLL & FCKSM_CSDCMD & OSCIOFNC_OFF & POSCMOD_NONE)
    _CONFIG1(JTAGEN_OFF & GCP_OFF & GWRP_OFF & COE_OFF & ICS_PGx1 & FWDTEN_OFF)

    
	#define BYTESPERSECTOR 137
	#define SECTORSPERTRACK 32
	#define NUMBEROFDRIVES 16
	
	// Read from buffer 1, which is IO ports 11Q and 10Q
	union {
		struct {
		  unsigned W:1;			//Enter Write Data
		  unsigned M:1;			//Head Move allowed
		  unsigned H:1;			//Head loaded
		  unsigned :2;
		  unsigned I:1;			//Interrupt enabled
		  unsigned Z:1;			//At track zero
		  unsigned R:1; 		//New Read Data Available
		  unsigned SectorTrue:1;
		  unsigned Sector:5;
		  unsigned :2;
		};
		unsigned int Word;
	} PMDRead1bits;

	// Write to buffer 1, IO ports 11Q and 10Q
	union {
		struct {
		  unsigned Drive:4;
		  unsigned :3;
		  unsigned Clear:1;
		  unsigned I:1;			//Step Head in
		  unsigned O:1;			//Step Head out
		  unsigned H:1;			//Load Head
		  unsigned U:1;			//Unload Head
		  unsigned E:1;			//Interrupt enable - Not used
		  unsigned D:1;			//Interrupt disable	- Not used
		  unsigned C:1;			//Head current - Not used
		  unsigned W:1;			//Write enable - start write sequence
		};
		unsigned int Word;
	}  PMDWrite1bits;
	  
	char Buffer[BYTESPERSECTOR + 8];
	int BufferIndex;
	unsigned int Track[NUMBEROFDRIVES];
	FSFILE * pFile[NUMBEROFDRIVES];
	int SelectedDrive = -1;
	char FileName[] = "DISK0.DSK";
	int Sector, NextSector;
	int i;
	int NeedToWrite = 0;
	int NeedToRead = 0;
	int NewSector = 0;
	long WriteAdd;
	int WriteDrive;

//Interrupt Handler
void _ISRFAST _PMPInterrupt(void)
{
	if (PMSTATbits.OB2E)	//Read from port 0x0A (12Q):
		PMDOUT2 = Buffer[BufferIndex++];
	else if (PMSTATbits.IB2F)	//Write to port 0x0A (12Q):
	{
		Buffer[BufferIndex++ - 1] = (char) (0x00FF & PMDIN2);
		if (!NeedToWrite)
		{
			WriteAdd = BYTESPERSECTOR * ((long)(Sector + SECTORSPERTRACK * Track[SelectedDrive]));
			WriteDrive = SelectedDrive;
			NeedToWrite = 1;	//need to write to disk
		}
	}
	else if (PMSTATbits.OB1E)	//Read from port 0x09 (11Q):
	{
		if (!PMDRead1bits.SectorTrue)
		{
			PMDRead1bits.SectorTrue = 1;	//not ready to change sector
			PMDRead1bits.R = 1;		//not ready to read
			PMDRead1bits.W = 1;		//not ready to write
		}
		PMDOUT1 = PMDRead1bits.Word;
	}
	else if (PMSTATbits.IB1F)	//Write to port 0x09 (11Q):
	{
		if (!PMDRead1bits.M)
		{
			PMDRead1bits.M = 1;		//don't move head
		   	PMDOUT1 = PMDRead1bits.Word;
		}
		PMDWrite1bits.Word = PMDIN1;
	}
	else if (PMSTATbits.OB0E)	//Read from port 0x08 (10Q):
	{
		if (NewSector) NeedToRead = 1;
		PMDOUT1 = PMDRead1bits.Word;
	}
	else if (PMSTATbits.IB0F)	//Write to port 0x08 (10Q):  Clear and/or select drive
	{
	   	PMDWrite1bits.Word = PMDIN1;
		SelectedDrive = PMDWrite1bits.Drive;
	}
	_PMPIF = 0;	//clear interrupt flag
}


int main (void)
{
   	Sector = 0;
	NextSector = 0;
	SelectedDrive = 0;

	for (i = 0; i < NUMBEROFDRIVES; ++i)
	{
	   pFile[i] = 0;
	   Track[i] = 0;
	}


//initalize
	CLKDIVbits.RCDIV = 0;	// 32MHz clock

//Turn off analog
   	AD1PCFGL = 0xFFFF;	//Turn off analog

//Set up PPS for SPI pins
	iPPSInput(IN_FN_PPS_SDI1,IN_PIN_PPS_RP8);    //SPI data input
	iPPSOutput(OUT_PIN_PPS_RP6,OUT_FN_PPS_SCK1OUT);    //SPI clock output
	iPPSOutput(OUT_PIN_PPS_RP7,OUT_FN_PPS_SDO1);    //SPI data output

//Set pullups
	EnablePullUpCN6;	//Pullup for CD
	EnablePullUpCN7;	//Pullup for WP
	
	PMMODEbits.MODE = 1;	//enhanced PSP
	_IRQM = 1;	//interrupt on read or write to PMP
	_PMPIF = 0;	//clear interrupt flag
	_PMPIE = 1;	//enable interrupts
	PMCONbits.PMPEN = 1; //enabled

	PMDRead1bits.Word = 0x0000;		//disk drive enabled
	PMDOUT1 = PMDRead1bits.Word;
	PMDOUT2 = 0;

	while (!MDD_MediaDetect());
// Initialize the library
	while (!FSInit());

	pFile[SelectedDrive] = FSfopen(FileName, "r+");
	FileName[4] = '1';
	pFile[1] = FSfopen(FileName, "r+");
	FileName[4] = '2';
	pFile[2] = FSfopen(FileName, "r+");
	FileName[4] = '3';
	pFile[3] = FSfopen(FileName, "r+");


	while(1)	//main loop
	{
		if (PMDRead1bits.SectorTrue)
		{
			Sector = NextSector;
			if (++NextSector >= SECTORSPERTRACK) NextSector = 0;
		   	PMDRead1bits.Sector = NextSector;
			PMDRead1bits.SectorTrue = 0;	//OK to read next sector
			NewSector = 1;
			PMDOUT1 = PMDRead1bits.Word;
		}
		
		if (NeedToRead)
		{	
			if (NeedToWrite)	//need to write data
			{
				FSfseek(pFile[WriteDrive], WriteAdd, SEEK_SET);
				FSfwrite(Buffer, 1, BYTESPERSECTOR, pFile[WriteDrive]);
				NeedToWrite = 0;
			}
			FSfseek(pFile[SelectedDrive], BYTESPERSECTOR * ((long)(Sector + SECTORSPERTRACK * Track[SelectedDrive])), SEEK_SET);
			FSfread(Buffer, 1, BYTESPERSECTOR, pFile[SelectedDrive]);
			if (!PMDRead1bits.SectorTrue && !PMDRead1bits.M)
			{
				NewSector = 0;
				NeedToRead = 0;
			    PMDOUT2 = Buffer[0];
			    BufferIndex = 1;
			    PMDRead1bits.R = 0;
				PMDRead1bits.W = 0;
				PMDOUT1 = PMDRead1bits.Word;
			}
		}


		if (PMDRead1bits.M)	
		{ 	
			if (PMDWrite1bits.I)
		   		++Track[SelectedDrive];
		   	if (PMDWrite1bits.O)
			   	if (Track[SelectedDrive] != 0)
			      	 --Track[SelectedDrive];
		   	if (Track[SelectedDrive] == 0)
		   		PMDRead1bits.Z = 0;
		   	else
		   		PMDRead1bits.Z = 1;
			PMDRead1bits.M = 0;		//OK to move head
		   	PMDOUT1 = PMDRead1bits.Word;
		}
	}
}
